from logging import Formatter
from copy import copy

from circleguard.enums import Mod


########### LOGGING ##############

TRACE = 5

# Colored logs adapted from
# https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
COLOR_MAPPING = {
    "TRACE"    : 90, # bright black
    "DEBUG"    : 94, # bright blue
    "INFO"     : 95, # bright magenta
    "WARNING"  : 31, # red
    "ERROR"    : 91, # bright red
    "CRITICAL" : 41, # white on red bg

    "NAME"     : 32,  # green
    "MESSAGE"  : 93,  # bright yellow
    "FILENAME" : 92,  # bright green
    "LINENO"   : 91   # bright red
}

COLOR_PREFIX = '\033['
COLOR_SUFFIX = '\033[0m'

class ColoredFormatter(Formatter):

    def __init__(self, patern):
        Formatter.__init__(self, patern)

    def format(self, record):
        # c as in colored, not as in copy
        c_record = copy(record)

        # logging's choice of camelCase, not mine
        threadName = c_record.threadName
        color = COLOR_MAPPING["NAME"]
        c_threadName = ('{0}{1}m{2}{3}').format(COLOR_PREFIX, color, threadName, COLOR_SUFFIX)

        levelname = c_record.levelname
        color = COLOR_MAPPING.get(levelname, 37) # default white
        c_levelname = ('{0}{1}m{2}{3}').format(COLOR_PREFIX, color, levelname, COLOR_SUFFIX)

        name = c_record.name
        color = COLOR_MAPPING["NAME"]
        c_name = ('{0}{1}m{2}{3}').format(COLOR_PREFIX, color, name, COLOR_SUFFIX)

        message = c_record.msg # why is this msg, but we format it as %(message)s in the formatter? mysteries of life.
        color = COLOR_MAPPING["MESSAGE"]
        c_msg = ('{0}{1}m{2}{3}').format(COLOR_PREFIX, color, message, COLOR_SUFFIX)

        filename = c_record.filename
        color = COLOR_MAPPING["FILENAME"]
        c_filename = ('{0}{1}m{2}{3}').format(COLOR_PREFIX, color, filename, COLOR_SUFFIX)

        lineno = c_record.lineno
        color = COLOR_MAPPING["LINENO"]
        c_lineno = ('{0}{1}m{2}{3}').format(COLOR_PREFIX, color, lineno, COLOR_SUFFIX)

        c_record.threadName = c_threadName
        c_record.levelname = c_levelname
        c_record.name = c_name
        c_record.msg = c_msg
        c_record.filename = c_filename
        c_record.lineno = c_lineno

        return Formatter.format(self, c_record)

######### UTIL METHODS ###########

def mod_to_int(mod):
    """
    Returns the integer representation of a mod string. The mods in the string can be in any order -
    "HDDT" will be parsed the same as "DTHD".

    Args:
        String mod: The modstring to convert.

    Returns:
        The integer representation of the passed mod string.

    """

    mod_total = 0
    for acronym in [mod[i:i+2] for i in range(0, len(mod), 2)]:
        mod_total += Mod[acronym].value

    return mod_total

# https://github.com/kszlim/osu-replay-parser/blob/master/osrparse/replay.py#L64
def bits(n):
    if n == 0:
        yield 0
    while n:
        b = n & (~n+1)
        yield b
        n ^= b


class Interpolation:
    """A utility class containing coordinate interpolations."""

    @staticmethod
    def linear(x1, x2, r):
        """
        Linearly interpolates coordinate tuples x1 and x2 with ratio r.

        Args:
            Float x1: The startpoint of the interpolation.
            Float x2: The endpoint of the interpolation.
            Float r: The ratio of the points to interpolate to.
        """

        return ((1 - r) * x1[0] + r * x2[0], (1 - r) * x1[1] + r * x2[1])

    @staticmethod
    def before(x1, x2, r):
        """
        Returns the startpoint of the range.

        Args:
            Float x1: The startpoint of the interpolation.
            Float x2: The endpoint of the interpolation.
            Float r: Ignored.
        """

        return x2

def interpolate(data1, data2, interpolation=Interpolation.linear, unflip=False):
    """
    Interpolates the longer of the datas to match the timestamps of the shorter.

    Args:
    List data1: A list of tuples of (t, x, y).
        List data2: A list of tuples of (t, x, y).
        Boolean unflip: Preserves input order of data1 and data2 if True.

    Returns:
        If unflip:
            The tuple (data1, data2), where one is interpolated to the other
            and said other without uninterpolatable points.
        Else:
            The tuple (clean, inter), respectively the shortest of
            the datasets without uninterpolatable points and the longest
            interpolated to the timestamps of shortest.

    """

    flipped = False

    # if the first timestamp in data2 is before the first in data1 switch
    # so data1 always has some timestamps before data2.
    if data1[0][0] > data2[0][0]:
        flipped = not flipped
        (data1, data2) = (data2, data1)

    # get the smallest index of the timestamps after the first timestamp in data2.
    i = next((i for (i, p) in enumerate(data1) if p[0] > data2[0][0]))

    # remove all earlier timestamps, if data1 is longer than data2 keep one more
    # so that the longest always starts before the shorter dataset.
    data1 = data1[i:] if len(data1) < len(data2) else data1[i - 1:]

    if len(data1) > len(data2):
        flipped = not flipped
        (data1, data2) = (data2, data1)

    # for each point in data1 interpolate the points around the timestamp in data2.
    j = 0
    inter = []
    clean = []
    for between in data1:
        # keep a clean version with only values that can be interpolated properly.
        clean.append(between)

        # move up to the last timestamp in data2 before the current timestamp.
        while j < len(data2) - 1 and data2[j][0] < between[0]:
            j += 1

        if j == len(data2) - 1:
            break

        before = data2[j]
        after = data2[j + 1]

        # calculate time differences
        # dt1 =  ---2       , data1
        # dt2 = 1-------3   , data2
        dt1 = between[0] - before[0]
        dt2 = after[0] - before[0]

        # skip trying to interpolate to this event
        # if its surrounding events are not set apart in time
        # and replace it with the event before it
        if dt2 == 0:
            inter.append((between[0], *before[1:]))
            continue

        # interpolate the coordinates in data2
        # according to the ratios of the time differences
        x_inter = interpolation(before[1:], after[1:], dt1 / dt2)

        # filter out interpolation artifacts which send outliers even further away
        if abs(x_inter[0]) > 600 or abs(x_inter[1]) > 600:
            inter.append(between)
            continue

        t_inter = between[0]

        inter.append((t_inter, *x_inter))

    if unflip and flipped:
        (clean, inter) = (inter, clean)

    return (clean, inter)

def resample(timestamped, frequency):
    """
    Resample timestamped data at the given frequency.

    Args:
        List timestamped: A list of tuples of (t, x, y).
        Float frequency: The frequency to resample data to in Hz.

    Returns
        A list of tuples of (t, x, y) with constant time interval 1 / frequency.
    """

    i = 0
    t = timestamped[0][0]
    t_max = timestamped[-1][0]

    resampled = []

    while t < t_max:
        while timestamped[i][0] < t:
            i += 1

        dt1 = t - timestamped[i - 1][0]
        dt2 = timestamped[i][0] - timestamped[i - 1][0]

        inter = Interpolation.linear(timestamped[i - 1][1:], timestamped[i][1:], dt1 / dt2)

        resampled.append((t, *inter))
        t += 1000 / frequency

    return resampled



from circleguard.enums import Detect

steal_thresh = 18
rx_thresh = 50
num = 50
stddevs = None
cache = False
failfast = False

def include_predicate(replay):
    return True

include = include_predicate
detect = Detect.ALL



import abc
import logging
from typing import Iterable

import circleparse
import numpy as np

from circleguard.enums import Detect, RatelimitWeight
from circleguard import config
from circleguard.utils import TRACE


class Replay(abc.ABC):
    def __init__(self, timestamp, map_id, username, user_id, mods, replay_id, replay_data, detect, weight):
        """
        Initializes a Replay instance.

        Args:
            Datetime timestamp: When this replay was played.
            Integer map_id: The map id the replay was played on, or 0 if unknown or on an unsubmitted map.
            String username: The username of the player who made the replay.
            Integer user_id: The id of the player who made the replay, or 0 if unknown.
            Integer mods: The mods the replay was played with.
            Integer replay_id: The id of this replay, or 0 if it does not have an id (unsubmitted replays have no id).
            List [circleparse.Replay.ReplayEvent] replay_data: An array containing objects with the attributes x, y, time_since_previous_action,
                            and keys_pressed. If the replay could not be loaded (from the api or otherwise), this field should be None.
                            This means that this replay will not be compared against other replays or investigated for cheats.
            Detect detect: What cheats to run tests to detect.
            RatelimitWeight weight: How much it 'costs' to load this replay from the api. If the load method of the replay makes no api calls,
                            this value is RatelimitWeight.NONE. If it makes only light api calls (anything but get_replay), this value is
                            RatelimitWeight.LIGHT. If it makes any heavy api calls (get_replay), this value is RatelimitWeight.HEAVY.
                            See the RatelimitWeight documentation for more details.
        """

        self.timestamp = timestamp
        self.map_id = map_id
        self.username = username
        self.user_id = user_id
        self.mods = mods
        self.replay_id = replay_id
        self.replay_data = replay_data
        self.detect = detect
        self.weight = weight
        self.loaded = True

    def __repr__(self):
        return (f"Replay(timestamp={self.timestamp},map_id={self.map_id},user_id={self.user_id},mods={self.mods},detect={self.detect},"
               f"replay_id={self.replay_id},weight={self.weight},loaded={self.loaded},username={self.username})")

    def __str__(self):
        return f"Replay by {self.username} on {self.map_id}"

    @abc.abstractclassmethod
    def load(self, loader, cache):
        """
        Loads replay data of the replay, from the osu api or from some other source.
        Implementation is up to the specific subclass.

        To meet the specs of this method, subclasses must set replay.loaded to True after this method is called,
        and replay.replay_data must be a valid Replay object, as defined by circleparse.Replay. Both of these specs
        can be met if the superclass circleguard.Replay is initialized in the load method with a valid Replay, as
        circleguard.Replay.__init__ sets replay.loaded to true by default.
        """

        ...


    def as_list_with_timestamps(self):
        """
        Gets the playdata as a list of tuples of absolute time, x, y, and pressed keys.

        Returns:
            A list of tuples of (t, x, y, keys).
        """
        # get all offsets sum all offsets before it to get all absolute times
        timestamps = np.array([e.time_since_previous_action for e in self.replay_data])
        timestamps = timestamps.cumsum()

        # zip timestamps back to data and convert t, x, y, keys to tuples
        txyk = [[z[0], z[1].x, z[1].y, z[1].keys_pressed] for z in zip(timestamps, self.replay_data)]
        # sort to ensure time goes forward as you move through the data
        # in case someone decides to make time go backwards anyway
        txyk.sort(key=lambda p: p[0])
        return txyk

class ReplayMap(Replay):
    """
    Represents a Replay submitted to online servers, that can be retrieved from the osu api.

    The only things you need to know to instantiate a ReplayMap
    are the user who made the replay, and the map it was made on.

    Attributes:
        Integer map_id: The id of the map the replay was made on.
        Integer user_id: The id of the user who made the replay.
        Integer mods: The mods the replay was played with. None if not set when instantiated and has not been loaded yet -
                      otherwise, set to the mods the replay was made with.
        String username: A readable representation of the user who made the replay. If passed,
                         username will be set to this string. Otherwise, it will be set to the user id.
                         This is so you don't need to know a user's username when creating a ReplayMap, only their id.
                         However, if the username is known (by retrieving it through the api, or other means), it is better
                         to represent the Replay with a player's name than an id. Both username and user_id will
                         obviously still be available to you through the result object after comparison.
        Detect detect: What cheats to run tests to detect.
        Boolean loaded: Whether this replay has been loaded. If True, calls to #load will have no effect.
                        See #load for more information.
        RatelimitWeight weight: RatelimitWeight.HEAVY, as this class' load method makes a heavy api call. See RatelimitWeight
                                documentation for more information.
    """

    def __init__(self, map_id, user_id, mods=None, username=None, detect=None):
        """
        Initializes a ReplayMap instance.

        Args:
            Integer map_id: The id of the map the replay was made on.
            Integer user_id: The id of the user who made the replay.
            Integer mods: The mods the replay was played with. If this is not set, the top scoring replay of the user on the
                          given map will be loaded. Otherwise, the replay with the given mods will be loaded.
            String username: A readable representation of the user who made the replay. If passed,
                             username will be set to this string. Otherwise, it will be set to the user id.
                             This is so you don't need to know a user's username when creating a ReplayMap, only their id.
                             However, if the username is known (by retrieving it through the api, or other means), it is
                             better to represent the Replay with a player's name than an id. Both username and user_id
                             will obviously still be available to you through the result object after comparison.
            Detect detect: What cheats to run tests to detect.
        """

        self.log = logging.getLogger(__name__ + ".ReplayMap")
        self.map_id = map_id
        self.user_id = user_id
        self.mods = mods
        self.detect = detect if detect is not None else config.detect
        self.weight = RatelimitWeight.HEAVY
        self.loaded = False
        self.username = username if username else user_id

    def __repr__(self):
        if self.loaded:
            return(f"ReplayMap(timestamp={self.timestamp},map_id={self.map_id},user_id={self.user_id},mods={self.mods},"
                f"detect={self.detect},replay_id={self.replay_id},weight={self.weight},loaded={self.loaded},"
                f"username={self.username})")

        else:
            return (f"ReplayMap(map_id={self.map_id},user_id={self.user_id},mods={self.mods},detect={self.detect},"
                f"weight={self.weight},loaded={self.loaded},username={self.username})")

    def __str__(self):
        return f"{'Loaded' if self.loaded else 'Unloaded'} ReplayMap by {self.username} on {self.map_id}"

    def load(self, loader, cache=None):
        """
        Loads the data for this replay from the api. This method silently returns if replay.loaded is True.

        The superclass Replay is initialized after this call, setting replay.loaded to True. Multiple
        calls to this method will have no effect beyond the first.
        """
        self.log.debug("Loading %r", self)
        if(self.loaded):
            self.log.debug("%s already loaded, not loading", self)
            return
        info = loader.user_info(self.map_id, user_id=self.user_id, mods=self.mods)
        replay_data = loader.replay_data(info, cache=cache)
        Replay.__init__(self, info.timestamp, self.map_id, self.username, self.user_id, info.mods, info.replay_id, replay_data, self.detect, self.weight)
        self.log.log(TRACE, "Finished loading %s", self)


class ReplayPath(Replay):
    """
    Represents a Replay saved locally.

    To instantiate a ReplayPath, you only need to know the path to the osr file. Although this class still
    loads some information from the api - like the map id and the user id - no RatelimitWeight.HEAVY calls
    are made, making this a relatively light replay to load. Of course, the replay has to already be downloaded
    to instantiate this class, sometimes making it less than ideal.

    Attributes:
        [String or Path] path: A pathlike object representing the absolute path to the osr file.
        Detect detect: What cheats to run tests to detect.
        Boolean loaded: Whether this replay has been loaded. If True, calls to #load will have no effect.
                        See #load for more information.
        RatelimitWeight weight: RatelimitWeight.LIGHT, as this class' load method makes only light api calls.
                                See RatelimitWeight documentation for more information.
    """

    def __init__(self, path, detect=None):
        """
        Initializes a ReplayPath instance.

        Args:
            [String or Path] path: A pathlike object representing the absolute path to the osr file.
            Detect detect: What cheats to run tests to detect.
        """

        self.log = logging.getLogger(__name__ + ".ReplayPath")
        self.path = path
        self.detect = detect if detect is not None else config.detect
        self.weight = RatelimitWeight.LIGHT
        self.loaded = False

    def __repr__(self):
        if self.loaded:
            return (f"ReplayPath(path={self.path},map_id={self.map_id},mods={self.mods},detect={self.detect},"
                    f"replay_id={self.replay_id},weight={self.weight},loaded={self.loaded},username={self.username})")
        else:
            return f"ReplayPath(path={self.path},detect={self.detect},weight={self.weight},loaded={self.loaded})"

    def __str__(self):
        if self.loaded:
            return f"Loaded ReplayPath by {self.username} on {self.map_id} at {self.path}"
        else:
            return f"Unloaded ReplayPath at {self.path}"

    def load(self, loader, cache=None):
        """
        Loads the data for this replay from the osr file given by the path. See circleparse.parse_replay_file for
        implementation details. This method has no effect if replay.loaded is True.

        The superclass Replay is initialized after this call, setting replay.loaded to True. Multiple
        calls to this method will have no effect beyond the first.

        The cache argument here currently has no effect, and is only added for homogeneity with ReplayMap#load.
        """

        self.log.debug("Loading ReplayPath %r", self)
        if self.loaded:
            self.log.debug("%s already loaded, not loading", self)
            return

        loaded = circleparse.parse_replay_file(self.path)
        map_id = loader.map_id(loaded.beatmap_hash)
        user_id = loader.user_id(loaded.player_name)

        Replay.__init__(self, loaded.timestamp, map_id, loaded.player_name, user_id, loaded.mod_combination,
                        loaded.replay_id, loaded.play_data, self.detect, self.weight)
        self.log.log(TRACE, "Finished loading %s", self)


class Check():
    """
    Contains a list of Replay objects (or subclasses thereof) and how to proceed when
    investigating them for cheats.

    Attributes:
        List [Replay] replays: A list of Replay objects.
        List [Replay] replays2: A list of Replay objects to compare against 'replays' if passed.
        Integer steal_thresh: If a comparison scores below this value, its Result object has ischeat set to True.
                Defaults to 18, or the config value if changed.
        Integer rx_thresh: if a replay has a ur below this value, it is considered cheated.
                Deaults to 50, or the config value if changed.
        Boolean cache: Whether to cache the loaded replays. Defaults to False, or the config value if changed.
        String mode: "single" if only replays was passed, or "double" if both replays and replays2 were passed.
        Boolean loaded: False at instantiation, set to True once check#load is called. See check#load for
                more details.
        Detect detect: What cheats to run tests to detect.
    """

    def __init__(self, replays, replays2=None, cache=None, steal_thresh=None, rx_thresh=None, include=None, detect=None):
        """
        Initializes a Check instance.

        If only replays is passed, the replays in that list are compared with themselves. If
        both replays and replays2 are passed, the replays in replays are compared only with the
        replays in replays2. See comparer#compare for a more detailed description.

        Args:
            List [Replay] replays: A list of Replay objects.
            List [Replay] replays2: A list of Replay objects to compare against 'replays' if passed.
            Boolean cache: Whether to cache the loaded replays. Defaults to False, or the config value if changed.
            Integer steal_thresh: If a Comparison scores below this value, it is considered cheated.
                    Defaults to 18, or the config value if changed.
            Integer rx_thresh: if a replay has a ur below this value, it is considered cheated.
                    Deaults to 50, or the config value if changed.
            Function include: A Predicate function that returns True if the replay should be loaded, and False otherwise.
                    The include function will be passed a single argument - the circleguard.Replay object, or one
                    of its subclasses.
            Detect detect: What cheats to run tests to detect. This will only overwrite replay's settings in this Check
                    if the replays were not given a Detect different from the (default) config value.
        """

        self.log = logging.getLogger(__name__ + ".Check")
        self.replays = replays # list of ReplayMap and ReplayPath objects, not yet processed
        self.replays2 = replays2 if replays2 else [] # make replays2 fake iterable, for #filter mostly
        self.detect = detect if detect else config.detect
        for r in self.all_replays():
            # if detect was not passed to Replays they default to config.detect,
            # we should only overwrite when detect wasn't explicitly passed to
            # the replay
            if r.detect == config.detect:
                r.detect = self.detect
        self.mode = "double" if replays2 else "single"
        self.loaded = False
        self.steal_thresh = steal_thresh if steal_thresh else config.steal_thresh
        self.rx_thresh = rx_thresh if rx_thresh else config.rx_thresh
        self.cache = cache if cache else config.cache
        self.include = include if include else config.include

    def filter(self):
        """
        Filters self.replays and self.replays2 to contain only Replays where self.include returns True
        when the Replay is passed. This gives total control to what replays end up getting loaded
        and compared.
        """

        self.log.info("Filtering replays from Check")
        self.replays = [replay for replay in self.replays if self._include(replay)]
        self.replays2 = [replay for replay in self.replays2 if self._include(replay)]


    def _include(self, replay):
        """
        An internal helper method to create log statements from inside a list comprehension.
        """

        if self.include(replay):
            self.log.log(TRACE, "%r passed include(), keeping in Check replays", replay)
            return True
        else:
            self.log.debug("%r failed include(), filtering from Check replays", replay)
            return False

    def load(self, loader):
        """
        If check.loaded is already true, this method silently returns. Otherwise, loads replay data for every
        replay in both replays and replays2, and sets check.loaded to True. How replays are loaded is up to
        the implementation of the specific subclass of the Replay. Although the subclass may not use the loader
        object, it is still passed regardless to reduce type checking. For implementation details, see the load
        method of each Replay subclass.

        Args:
            Loader loader: The loader to handle api requests, if required by the Replay.
        """

        self.log.info("Loading replays from Check")

        if self.loaded :
            self.log.debug("Check already loaded, not loading individual Replays")
            return
        for replay in self.replays:
            replay.load(loader, self.cache)
        if self.replays2:
            for replay in self.replays2:
                replay.load(loader, self.cache)
        self.loaded = True
        self.log.debug("Finished loading Check object")

    def all_replays(self) -> Iterable[Replay]:
        """
        Convenience method for accessing all replays stored in this object.

        Returns:
            A list of all replays in this Check object (replays1 + replays2)
        """
        return self.replays + self.replays2



import sqlite3
import logging
import os

import wtc

from circleguard.loader import Loader
from circleguard.exceptions import CircleguardException
from circleguard.utils import TRACE


class Cacher:
    """
    Handles compressing and caching replay data to a database.

    Each Cacher instance maintains its own database connection.
    Be wary of instantiating too many.
    """

    def __init__(self, cache, path):
        """
        Initializes a Cacher instance.

        Args:
            Boolean cache: Whether replays should be cached or not.
            Path path: A pathlike object representing the absolute path to the database
        """

        self.log = logging.getLogger(__name__)
        self.should_cache = cache
        if not os.path.isfile(str(path)):
            self._create_cache(str(path))
        self.conn = sqlite3.connect(str(path))
        self.cursor = self.conn.cursor()

    def cache(self, lzma_bytes, user_info, should_cache=None):
        """
        Writes the given lzma bytes to the database, linking it to the given map and user.
        If an entry with the given map_id and user_id already exists, it is overwritten with
        the given lzma_bytes (after compression) and replay_id.

        The lzma string is compressed with wtc compression. See Cacher.compress and wtc.compress for more.

        A call to this method has no effect if the Cacher's should_cache is False.

        Args:
            String map_id: The map id to insert into the db.
            Bytes lzma_bytes: The lzma bytes to compress and insert into the db.
            UserInfo user_info: The UserInfo object representing this replay.
            Boolean should_cache: If this is passed, overwrites the option set at initialization time.
        """

        self.log.debug("Caching lzma bytes")
        should_cache = should_cache if should_cache else self.should_cache

        if(not should_cache):
            self.log.debug("should_cache is false, not caching")
            return
        compressed_bytes = self._compress(lzma_bytes)

        map_id = user_info.map_id
        user_id = user_info.user_id
        mods = user_info.mods
        replay_id = user_info.replay_id

        result = self.cursor.execute("SELECT COUNT(1) FROM replays WHERE map_id=? AND user_id=? AND mods=?", [map_id, user_id, mods]).fetchone()[0]
        self.log.log(TRACE, "Writing compressed lzma to db")
        if(result): # already exists so we overwrite (this happens when we call Cacher.revalidate)
            self._write("UPDATE replays SET replay_data=?, replay_id=? WHERE map_id=? AND user_id=? AND mods=?", [compressed_bytes, replay_id, map_id, user_id, mods])
        else: # else just insert
            self._write("INSERT INTO replays VALUES(?, ?, ?, ?, ?)", [map_id, user_id, compressed_bytes, replay_id, mods])

    def revalidate(self, loader, user_info):
        """
        Revalidates every entry in user_info, which may contain different maps or users. If an entry exists in the cache with a UserInfo's
        map_id, user_id, and enabled_mods, the score is redownloaded and the outdated score is replaced with the new one in the cache.

        Args:
            Loader loader: The Loader from the circleguard instance to redownload beatmaps with if they are outdated.
            List [UserInfo]: A list of UserInfo objects containing the up-to-date information of user's replays.
        """

        self.log.info("Revalidating cache with %d user_infos", len(user_info))

        for info in user_info:
            map_id = info.map_id
            user_id = info.user_id
            mods = info.enabled_mods

            self.log.log(TRACE, "Revalidating entry with map id %s, user %d, mods %s", map_id, user_id, mods)

            result = self.cursor.execute("SELECT replay_id FROM replays WHERE map_id=? AND user_id=? AND mods=?", [map_id, user_id, mods]).fetchall()
            if(not result):
                self.log.trace("Nothing cached with map id %s, user %d, mods %s", map_id, user_id, mods)
                continue # nothing cached

            db_replay_id = result[0][0] # blame sqlite for nesting tuples in lists
            new_replay_id = info.replay_id

            if(db_replay_id != new_replay_id):
                if(db_replay_id > new_replay_id):
                    raise CircleguardException("The cached replay id of {} is higher than the new replay id of {}. Map id: {}, User id: {}, mods: {}"
                                                .format(db_replay_id, new_replay_id, user_id, map_id, mods))

                self.log.info("Cached replay on map %d by user %d with mods %d is outdated, redownloading", map_id, user_id, mods)
                lzma_data = loader.replay_data(info)
                if(lzma_data is None):
                    raise CircleguardException("We could not load lzma data for map {}, user {}, mods {}, replay available {} while revalidating."
                                                .format(map_id, user_id, mods, info.replay_available))
                self.cache(lzma_data, info)

    def check_cache(self, map_id, user_id, mods):
        """
        Checks if a replay exists on the given map_id by the given user_id with the given mods, and returns the decompressed wtc (equivelant to an lzma) string if so.

        Args:
            String map_id: The map_id to check for.
            String user_id: The user_id to check for.
            Integer mods: The bitwise enabled mods for the play.

        Returns:
            The lzma bytes that would have been returned by decoding the base64 api response, or None if it wasn't cached.
        """
        self.log.log(TRACE, "Checking cache for a replay on map %d by user %d with mods %s", map_id, user_id, mods)
        result = self.cursor.execute("SELECT replay_data FROM replays WHERE map_id=? AND user_id=? AND mods=?", [map_id, user_id, mods]).fetchone()
        if(result):
            self.log.debug("Loading replay on map %d by user %d with mods %s from cache", map_id, user_id, mods)
            return wtc.decompress(result[0])
        self.log.log(TRACE, "No replay found in cache")
        return None

    def _write(self, statement, args):
        """
        Writes an sql statement with the given args to the databse.

        Args:
            String statement: The prepared sql statement to execute.
            List args: The values to insert into the prepared sql statement.
                       Must be of length equal to the number of missing values in the statement.
        """

        self.cursor.execute(statement, args)
        self.conn.commit()

    def _compress(self, lzma_bytes):
        """
        Compresses the lzma string to a (smaller) wtc string to store in the database.

        Args:
            Bytes lzma_bytes: The lzma bytes, returned by the api for replays, to compress.

        Returns:
            A compressed bytes from the given bytes, using lossy wtc compression.
        """

        self.log.log(TRACE, "Compressing lzma bytes")
        return wtc.compress(lzma_bytes)

    def _create_cache(self, path):
        self.log.info("Cache not found at path %s, creating cache", path)
        if not os.path.exists(os.path.split(path)[0]):  # create dir if nonexistent
            os.makedirs(os.path.split(path)[0])
        conn = sqlite3.connect(str(path))
        c = conn.cursor()
        c.execute("""CREATE TABLE "REPLAYS"(
            "MAP_ID" INTEGER NOT NULL,
            "USER_ID" INTEGER NOT NULL,
            "REPLAY_DATA" MEDIUMTEXT NOT NULL,
            "REPLAY_ID" INTEGER NOT NULL,
            "MODS" INTEGER NOT NULL,
            PRIMARY KEY("REPLAY_ID")
        )""")
        conn.close()



import circleguard.utils as utils
from circleguard.enums import Mod

class UserInfo():
    """
    A container class, holding a user's map id, user id, username, replay id, enabled mods, and replay availability for a given replay.
    """

    def __init__(self, timestamp, map_id, user_id, username, replay_id, mods, replay_available):
        """
        Initializes a UserInfo class.

        Args:
            Datetime timestamp: When this replay was set.
            Integer map_id: The id of the map the replay was made on.
            Integer user_id: The id of the player who set the replay.
            String username: The username of the player who set the replay.
            Integer replay_id: The id of the replay.
            Integer enabled_mods: The bitwise mod combination of the mods set on the replay.
            Boolean replay_available: Whether this replay is available from the api or not.
        """

        self.timestamp = timestamp
        self.map_id = map_id
        self.user_id = user_id
        self.username = username
        self.replay_id = replay_id
        self.mods = mods
        self.replay_available = replay_available



from datetime import datetime
import time
import base64
import logging
from lzma import LZMAError

import requests
from requests import RequestException
import circleparse
import ossapi

from circleguard.user_info import UserInfo
from circleguard.enums import Error
from circleguard.exceptions import (InvalidArgumentsException, APIException, CircleguardException,
                        RatelimitException, InvalidKeyException, ReplayUnavailableException, UnknownAPIException,
                        InvalidJSONException)
from circleguard.utils import TRACE

def request(function):
    """
    Decorator intended to appropriately handle all request and api related exceptions.

    Also checks if we can refresh the time at which we started our requests because
    it's been more than RATELIMIT_RESET since the first request of the cycle.

    If we've refreshed our ratelimits, sets start_time to be the current datetime.
    """

    def wrapper(*args, **kwargs):
        difference = datetime.now() - Loader.start_time
        if(difference.seconds > Loader.RATELIMIT_RESET):
            Loader.start_time = datetime.now()
        # catch them exceptions boy
        ret = None
        self = args[0]
        try:
            ret = function(*args, **kwargs)
        except RatelimitException:
            self._enforce_ratelimit()
            # wrap function with the decorator then call decorator
            ret = request(function)(*args, **kwargs)
        except InvalidJSONException as e:
            self.log.warning("Invalid json exception: {}. API likely having issues; sleeping for 3 seconds then retrying".format(e))
            time.sleep(3)
            ret = request(function)(*args, **kwargs)
        except InvalidKeyException as e:
            raise CircleguardException("The given key is invalid")
        except RequestException as e:
            self.log.warning("Request exception: {}. Likely a network issue; sleeping for 5 seconds then retrying".format(e))
            time.sleep(5)
            ret = request(function)(*args, **kwargs)
        except ReplayUnavailableException as e:
            self.log.warning("We expected a replay from the api, but it was unable to deliver it: {}".format(e))
            ret = None
        return ret
    return wrapper


def check_cache(function):
    """
    Decorator that checks if the replay by the given user_id on the given map_id is already
    cached. If so, returns a Replay instance from the cached data instead of requesting it
    from the api. Otherwise, it calls the function as normal.

    Note that self and user_info must be the first and second arguments to
    the function respectively.

    Returns:
        A Replay instance from the cached replay if it was cached,
        or the return value of the function if not.
    """

    def wrapper(*args, **kwargs):
        self = args[0]
        user_info = args[1]

        if self.cacher is None:
            return function(*args, **kwargs)

        lzma = self.cacher.check_cache(user_info.map_id, user_info.user_id, user_info.mods)
        if lzma :
            replay_data = circleparse.parse_replay(lzma, pure_lzma=True).play_data
            return replay_data
        else:
            return function(*args, **kwargs)
    return wrapper

class Loader():
    """
    Manages interactions with the osu api, using the ossapi wrapper.

    if the api ratelimits the key, we wait until we refresh our ratelimits and retry
    the request. Because the api does not provide the time until the next refresh (and we
    do not periodically retry the key), if the key is ratelimited outside of this class,
    the class may wait more time than necessary for the key to refresh.
    """

    RATELIMIT_RESET = 60 # time in seconds until the api refreshes our ratelimits
    start_time = datetime.min # when we started our requests cycle


    def __init__(self, key, cacher=None):
        """
        Initializes a Loader instance. If a Cacher is not provided,
        scores will not be loaded from cache or cached to the databse.
        """

        self.log = logging.getLogger(__name__)
        self.api = ossapi.ossapi(key)
        self.cacher = cacher

    @request
    def get_beatmap(self, map_id):
        """
        Returns the content of the beatmap of the given map. This request is
        not ratelimited and does not require an api key. Because of this, treat
        this endpoint with caution - the osu-tools repository uses this endpoint
        and peppy has said it is "ok to use for now", but even so it is not in
        the same category as other api endpoints.
        """
        return requests.get(f"https://osu.ppy.sh/osu/{map_id}").content

    @request
    def user_info(self, map_id, num=None, user_id=None, mods=None, limit=True):
        """
        Returns a list of UserInfo objects containing a user's
        (timestamp, map_id, user_id, username, replay_id, mods, replay_available)
        on the given map.

        If limit and user_id is set, it will return a single UserInfo object, not a list.

        Args:
            Integer map_id: The map id to get the replay_id from.
            Integer user_id: The user id to get the replay_id from.
            Boolean limit: If set, will only return a user's top score (top response). Otherwise, will
                          return every response (every score they set on that map under different mods)
            Integer mods: The mods the replay info to retieve were played with.
        """

        # we have to define a new variable to hold locals - otherwise when we call it twice inside the dict comprehension,
        # it rebinds to the comp scope and takes on different locals which is real bad. I spent many-a-hour figuring this out,
        # and if anyone has a more elegant solution I'm all ears.
        locals_ = locals()
        self.log.log(TRACE, "Loading user info on map %d with options %s",
                            map_id, {k: locals_[k] for k in locals_ if k != 'self'})

        if(num and (num > 100 or num < 1)):
            raise InvalidArgumentsException("The number of top plays to fetch must be between 1 and 100 inclusive!")

        if(not bool(user_id) ^ bool(num)):
            raise InvalidArgumentsException("One of either num or user_id must be passed, but not both")

        response = self.api.get_scores({"m": "0", "b": map_id, "limit": num, "u": user_id, "mods": mods})
        Loader.check_response(response)
        # yes, it's necessary to cast the str response to int before bool - all strings are truthy.
        # strptime format from https://github.com/ppy/osu-api/wiki#apiget_scores
        infos = [UserInfo(datetime.strptime(x["date"], "%Y-%m-%d %H:%M:%S"), map_id, int(x["user_id"]), str(x["username"]), int(x["score_id"]),
                          int(x["enabled_mods"]), bool(int(x["replay_available"]))) for x in response]

        return infos[0] if (limit and user_id) else infos # limit only applies if user_id was set


    @request
    def get_user_best(self, user_id, number):
        """
        Gets the top 100 best plays for the given user.

        Args:
            String user_id: The user id to get best plays of.
            Integer number: The number of top plays to retrieve. Must be between 1 and 100.

        Returns:
            A list of Integer map_ids for the given number of the user's top plays.

        Raises:
            InvalidArgumentsException if number is not between 1 and 100 inclusive.
        """

        self.log.log(TRACE, "Retrieving the best %d plays of user %d", number, user_id)
        if(number < 1 or number > 100):
            raise InvalidArgumentsException("The number of best user plays to fetch must be between 1 and 100 inclusive!")
        response = self.api.get_user_best({"m": "0", "u": user_id, "limit": number})
        Loader.check_response(response)

        return [int(x["beatmap_id"]) for x in response]


    @request
    def load_replay_data(self, map_id, user_id, mods=None):
        """
        Queries the api for replay data from the given user on the given map, with the given mods.

        Args:
            UserInfo user_info: The UserInfo representing this replay.
        Returns:
            The lzma bytes (b64 decoded response) returned by the api, or None if the replay was not available.

        Raises:
            CircleguardException if the loader instance has had a new session made yet.
            APIException if the api responds with an error we don't know.
        """

        self.log.log(TRACE, "Requesting replay data by user %d on map %d with mods %s", user_id, map_id, mods)
        response = self.api.get_replay({"m": "0", "b": map_id, "u": user_id, "mods": mods})
        Loader.check_response(response)

        return base64.b64decode(response["content"])

    @check_cache
    def replay_data(self, user_info, cache=None):
        """
        Loads the replay data specified by the user info.

        Args:
            UserInfo user_info: The UserInfo object representing this replay.

        Returns:
            The play_data field of an circleparse.Replay instance created from the replay data received from the api,
            or None if the replay was not available.

        Raises:
            UnknownAPIException if replay_available was 1, but we did not receive replay data from the api.
        """

        user_id = user_info.user_id
        map_id = user_info.map_id
        mods = user_info.mods
        if(not user_info.replay_available):
            self.log.debug("Replay data by user %d on map %d with mods %s not available", user_id, map_id, mods)
            return None

        lzma_bytes = self.load_replay_data(map_id, user_id, mods)
        if(lzma_bytes is None):
            raise UnknownAPIException("The api guaranteed there would be a replay available, but we did not receive any data. "
                                     "Please report this to the devs, who will open an issue on osu!api if necessary.")
        try:
            parsed_replay = circleparse.parse_replay(lzma_bytes, pure_lzma=True)
        # see https://github.com/circleguard/circlecore/issues/61
        # api sometimes returns corrupt replays
        except LZMAError:
            self.log.warning("lzma from %r could not be decompressed, api returned corrupt replay", user_info)
            return None
        replay_data = parsed_replay.play_data
        if self.cacher is not None:
            self.cacher.cache(lzma_bytes, user_info, should_cache=cache)
        return replay_data

    def map_id(self, map_hash):
        """
        Retrieves the corresponding map id for the given map_hash from the api.

        Returns:
            The corresponding map id, or 0 if the api returned no matches.
        """

        response = self.api.get_beatmaps({"h": map_hash})
        if response == []:
            return 0
        else:
            return int(response[0]["beatmap_id"])

    def user_id(self, username):
        """
        Retrieves the corresponding user id for the given username from the api.
        Note that the api currently has no method to keep track of name changes,
        meaning this method will return 0 for previous usernames of a user, rather
        than their true id.

        Returns:
            The corresponding user id, or 0 if the api returned no matches.
        """

        response = self.api.get_user({"u": username, "type": "string"})
        if response == []:
            return 0
        else:
            return int(response[0]["user_id"])

    @staticmethod
    def check_response(response):
        """
        Checks the given api response for any kind of error or unexpected response.

        Args:
            String response: The api-returned response to check.

        Raises:
            An Error corresponding to the type of error if there was an error. The mappings
            for the api error message and its corresponding error are in circleguard.enums.
        """

        if("error" in response):
            for error in Error:
                if(response["error"] == error.value[0]):
                    raise error.value[1](error.value[2])
            else:
                raise Error.UNKNOWN.value[1](Error.UNKNOWN.value[2]) # pylint: disable=unsubscriptable-object
                # pylint is dumb because Error is an enum and this is totally legal

    def _enforce_ratelimit(self):
        """
        Enforces the api ratelimit by sleeping the thread until it's safe to make requests again.
        """

        difference = datetime.now() - Loader.start_time
        seconds_passed = difference.seconds

        # sleep the remainder of the reset cycle so we guarantee it's been that long since the first request
        sleep_seconds = Loader.RATELIMIT_RESET - seconds_passed
        self._ratelimit(sleep_seconds)

    def _ratelimit(self, length):
        """
        Sleeps the thread for the specified amount of time. Called by #_enforce_ratelimit.

        Split into two functions mostly to allow Loader subclasses to overload a single function and easily
        get the time the Loader will be ratelimited for.
        """
        self.log.info("Ratelimited, sleeping for %s seconds.", length)
        time.sleep(length)



import logging

from circleguard.circleguard import Circleguard, set_options
from circleguard.replay import Check, Replay, ReplayMap, ReplayPath
from circleguard.enums import Detect
from circleguard.utils import TRACE, ColoredFormatter
from circleguard.loader import Loader
from circleguard.version import __version__

logging.addLevelName(TRACE, "TRACE")
formatter = ColoredFormatter("[%(threadName)s][%(name)s][%(levelname)s]  %(message)s  (%(filename)s:%(lineno)s)")
handler_stream = logging.StreamHandler()
handler_stream.setFormatter(formatter)
logging.getLogger("circleguard").addHandler(handler_stream)

__all__ = ["Circleguard", "set_options", "Check", "Replay", "ReplayMap",
           "ReplayPath", "Detect", "TRACE", "ColoredFormatter", "Loader",
           "__version__"]



from circleguard.replay import Replay
from circleguard.enums import ResultType

# Hierarchy
#                                 Result
#          InvestigationResult             ComparisonResult
#              RelaxResult               ReplayStealingResult
#
#
class Result():
    """
    The result of a test for cheats, either on a single replay or a
    collection of replays.

    Attributes:
        Boolean ischeat: Whetherone or more of the replays involved is cheated.
        ResultType type: What type of cheat test we are representing the results for.
    """

    def __init__(self, ischeat: bool, type_: ResultType):
        """
        Initializes a Result instance.

        Boolean ischeat: Whether one or more of the replays involved is cheated.
        ResultType type: What type of cheat test we are representing the results for.
        """
        self.ischeat = ischeat
        self.type = type_

class InvestigationResult(Result):
    """
    The result of a test for cheats on a single replay.

    Attributes:
        Replay replay: The replay investigated.
        Boolean ischeat: Whether the replay is cheated.
    """

    def __init__(self, replay: Replay, ischeat: bool, type_: ResultType):
        super().__init__(ischeat, type_)
        self.replay = replay

class ComparisonResult(Result):
    """
    The result of a test for cheats by comparing two replays.

    Attributes:
        Replay replay1: One of the replays involved.
        Replay replay2: The other replay involved.
        Boolean ischeat: Whether one of the replays is cheated.
    """

    def __init__(self, replay1: Replay, replay2: Replay, ischeat: bool, type_: ResultType):
        super().__init__(ischeat, type_)
        self.replay1 = replay1
        self.replay2 = replay2

class ReplayStealingResult(ComparisonResult):
    """
    The result of a test for replay stealing between two replays.

    This Result contains the two Replay objects that were compared, and the similarity
    between them. Attributes about the two replays (usernames, mods, map id, etc)
    can be accessed through the replay1 and replay2 objects, as well as the later_replay
    and earlier_replay objects, which are references to one of either replay1 or replay2.

    Attributes:
        Replay replay1: One of the replays involved.
        Replay replay2: The other replay involved.
        Replay earlier_replay: The earlier of the two replays (when the score was made). This is a
                reference to either replay1 or replay2.
        Replay later_replay: The later of the two replays (when the score was made). This is a
                reference to either replay1 or replay2.
        Integer similarity: How similar the two replays are (the lower, the more similar).
        Boolean ischeat: Whether one of the replays is cheated.
    """

    def __init__(self, replay1: Replay, replay2: Replay, similarity: int, ischeat: bool):
        super().__init__(replay1, replay2, ischeat, ResultType.REPLAY_STEALING)

        self.similarity = similarity
        if self.replay1.timestamp < self.replay2.timestamp:
            self.earlier_replay: Replay = self.replay1
            self.later_replay: Replay = self.replay2
        else:
            self.earlier_replay: Replay = self.replay2
            self.later_replay: Replay = self.replay1

class RelaxResult(InvestigationResult):
    """
    The result of a test for relax cheats.

    Attributes:
        Replay replay: The replay investigated.
        Integer ur: The unstable rate of the replay.
        Boolean ischeat: Whether the replay is cheated.
    """

    def __init__(self, replay: Replay, ur: int, ischeat: bool):
        super().__init__(replay, ischeat, ResultType.RELAX)
        self.ur = ur



import numpy as np
from circleguard.enums import Keys
from circleguard.result import RelaxResult
import circleguard.utils as utils

class Investigator:
    """
    A class for checking isolated replays for cheats.

    See Also:
        Comparer
    """

    def __init__(self, replay, beatmap, threshold):
        """
        Initializes an Investigator instance.

        Attributes:
            Replay replay: The Replay object to investigate.
            circleparse.Beatmap beatmap: The beatmap to calculate ur with.
            Integer threshold: If a replay has a lower ur than this value,
                    it is considered a cheted repaly.
        """

        self.replay = replay.as_list_with_timestamps()
        self.beatmap = beatmap
        self.threshold = threshold
        self.last_keys = [0, 0]

    def investigate(self):
        ur = self.ur()
        ischeat = True if ur < self.threshold else False
        yield RelaxResult(self.replay, ur, ischeat)

    def ur(self):
        hitobjs = self._parse_beatmap(self.beatmap)
        keypresses = self._parse_keys(self.replay)
        filtered_array = self._filter_hits(hitobjs, keypresses)
        diff_array = []

        for hit, press in filtered_array:
            diff_array.append(press[0]-hit[0])
        return np.std(diff_array) * 10

    def _parse_beatmap(self, beatmap):
        hitobjs = []

        # parse hitobj
        for hit in beatmap.hitobjects:
            hitobjs.append([hit.time, hit.x, hit.y])
        return hitobjs

    def _parse_keys(self, replay):
        keypresses = []
        self.last_keys = [0, 0]
        for keypress in replay:
            if self._check_keys(keypress[3]):
                    keypresses.append(keypress)
        return keypresses

    def _check_keys(self, pressed):
        checks = [pressed & key.value for key in (Keys.K1, Keys.K2)]
        if checks != self.last_keys and any(checks): 
            if not all(self.last_keys):  # skip if user was holding both buttons in previous event
                self.last_keys = checks
                return True
        self.last_keys = checks
        return False

    def _filter_hits(self, hitobjs, keypresses):
        array = []
        hitwindow = 150 + 50 * (5 - self.beatmap.difficulty["OverallDifficulty"]) / 5
        # inefficient as fugg, but not important rn
        # filters out all clicks that didn't hit anything and stores hit and hitobj to array
        for hitobj in hitobjs:
            temp_diffs = []
            temp_hits = []
            for press in keypresses:
                if hitobj[0] > press[0]-(hitwindow/2) and hitobj[0] < press[0]+(hitwindow/2):
                    diff = press[0]-hitobj[0]
                    temp_diffs.append(diff)
                    temp_hits.append(press)
            # only add one click per hitobj
            if len(temp_diffs) != 0:
                index = temp_diffs.index(min(temp_diffs, key=abs))
                array.append([hitobj, temp_hits[index]])
                keypresses.pop(keypresses.index(temp_hits[index]))  # remove used hits
        return array

class Hit:
    def __init__(self, hitobj, hit):
        self.x = hit[1]-hitobj[1]
        self.y = hit[2]-hitobj[2]
        self.error = hit[0]-hitobj[0]
        self.keys = [Keys(key_val) for key_val in utils.bits(hit[3])]
